/**
 * @module
 * @author Sébastien Règne
 * @license MIT
 */

/**
 * La classe d'un champ d'une expression <em>cron</em>.
 *
 * @class Field
 */
const Field = class {

    /**
     * Crée un champ d'une expression <em>cron</em> autorisant une seule valeur.
     *
     * @param {number} value La valeur autorisée pour le champ.
     * @returns {Field} Le champ avec la valeur.
     */
    static of(value) {
        return new Field([value]);
    }

    /**
     * Crée un champ d'une expression <em>cron</em> avec toutes les valeurs
     * autorisées (sans restriction).
     *
     * @param {number} min La valeur minimale (incluse) autorisée.
     * @param {number} max La valeur maximale (incluse) autorisée.
     * @returns {Field} Le champ avec toutes les valeurs (sans restriction).
     */
    static all(min, max) {
        const values = [];
        for (let value = min; value <= max; ++value) {
            values.push(value);
        }
        return new Field(values, false);
    }

    /**
     * Crée un champ d'une expression <em>cron</em> avec un intervalle de
     * valeurs autorisées.
     *
     * @param {number} min La valeur minimale (incluse) autorisée.
     * @param {number} max La valeur maximale (incluse) autorisée.
     * @param {number} step Le pas entre les valeurs.
     * @returns {Field} Le champ avec l'intervalle de valeurs.
     */
    static range(min, max, step) {
        const values = [];
        for (let value = min; value <= max; value += step) {
            values.push(value);
        }
        return new Field(values);
    }

    /**
     * Regroupe les valeurs de plusieurs champs.
     *
     * @param {Field[]} fields La liste des champs qui seront regroupés.
     * @returns {Field} Le champ avec toutes les valeurs.
     */
    static flat(fields) {
        return new Field(fields.flatMap((f) => f._values));
    }

    /**
     * La liste des valeurs autorisées pour le champ (en enlevant les doublons
     * et en triant).
     *
     * @type {number[]}
     * @private
     */
    _values;

    /**
     * <code>true</code> si le champ était différent de <code>"*"</code> ; sinon
     * <code>false</code>.
     *
     * @type {boolean}
     * @private
     */
    _restricted;

    /**
     * Crée un champ d'une expression <em>cron</em>.
     *
     * @param {number[]} values            La liste des valeurs autorisées pour
     *                                     le champ.
     * @param {boolean}  [restricted=true] <code>true</code> (par défaut) pour
     *                                     un champ qui était différent de
     *                                     <code>"*"</code> ; sinon
     *                                     <code>false</code>.
     * @private
     */
    constructor(values, restricted = true) {
        // Enlever les doublons des valeurs et trier les valeurs pour faciliter
        // les algorithmes.
        this._values = values.filter((v, i, a) => i === a.indexOf(v))
                             .sort((v1, v2) => v1 - v2);
        this._restricted = restricted;
    }

    /**
     * Retourne la marque indiquant si le champ était différent de
     * <code>"*"</code>.
     *
     * @returns {boolean} <code>true</code> si le champ était différent de
     *                    <code>"*"</code> ; sinon <code>false</code>.
     */
    get restricted() {
        return this._restricted;
    }

    /**
     * Retourne la valeur minimale.
     *
     * @returns {number} La valeur minimale.
     */
    get min() {
        return this._values[0];
    }

    /**
     * Retourne la valeur maximale.
     *
     * @returns {number} La valeur maximale.
     */
    get max() {
        return this._values[this._values.length - 1];
    }

    /**
     * Retourne la liste des valeurs autorisées pour le champ.
     *
     * @returns {number[]} La liste des valeurs.
     */
    values() {
        return this._values;
    }

    /**
     * Applique une fonction sur toutes les valeurs.
     *
     * @param {function(number, number, number[]): number} callback La fonction
     *                                                              appelée sur
     *                                                              les valeurs.
     * @returns {Field} Le nouveau champ avec les valeurs modifiées.
     */
    map(callback) {
        return new Field(this._values.map(callback), this._restricted);
    }

    /**
     * Teste si une valeur est présente dans la liste des valeurs.
     *
     * @param {number} value La valeur qui sera testée.
     * @returns {boolean} <code>true</code> si la valeur est présente ; sinon
     *                    <code>false</code>.
     */
    test(value) {
        return this._values.includes(value);
    }

    /**
     * Calcule la prochaine valeur supérieure à une valeur.
     *
     * @param {number} value La valeur initiale.
     * @returns {number|undefined} La prochaine valeur ; ou
     *                             <code>undefined</code> si la valeur initiale
     *                             est supérieure à la valeur maximale.
     */
    next(value) {
        return this._values.find((v) => value < v);
    }
};

/**
 * @module
 * @author Sébastien Règne
 * @license MIT
 */

/**
 * Les chaines spéciales avec leur équivalent.
 *
 * @type {Object<string, string>}
 * @private
 */
const NICKNAMES = {
    "@yearly":   "0 0 1 1 *",
    "@annually": "0 0 1 1 *",
    "@monthly":  "0 0 1 * *",
    "@weekly":   "0 0 * * 0",
    "@daily":    "0 0 * * *",
    "@midnight": "0 0 * * *",
    "@hourly":   "0 * * * *",
};

/* eslint-disable jsdoc/check-types -- Utiliser la notation Array<> car l'outil
 *     JSDoc ne gère pas les tableaux de types complexes déclarés avec [].
 *     https://github.com/jsdoc/jsdoc/issues/1133 */
/**
 * Les formes littérales des mois et des jours de la semaine. Les autres champs
 * (minute, heure et jour du mois) n'en ont pas. Ce sont les index des éléments
 * dans le tableau qui sont utilisés pour la correspondance.
 *
 * @type {Array<Array<?string>>}
 * @private
 */
const NAMES = [
    // Minute.
    [],
    // Heure.
    [],
    // Jour du mois.
    [],
    // Mois.
    [
        // Commencer par une valeur nulle pour que le mois de janvier soit à
        // l'index 1.
        null,
        "jan",
        "feb",
        "mar",
        "apr",
        "may",
        "jun",
        "jul",
        "aug",
        "sep",
        "oct",
        "nov",
        "dec",
    ],
    // Jour de la semaine.
    [
        "sun",
        "mon",
        "tue",
        "wed",
        "thu",
        "fri",
        "sat",
    ],
];
/* eslint-enable jsdoc/check-types */

/* eslint-disable jsdoc/check-types -- Utiliser la notation Array<> car l'outil
 *     JSDoc ne gère pas les tableaux de types complexes déclarés avec [].
 *     https://github.com/jsdoc/jsdoc/issues/1133 */
/**
 * Les valeurs minimales et maximales (incluses) saisissables dans les cinq
 * champs (pour des valeurs simples ou des intervalles).
 *
 * @type {Array<Object<string, number>>}
 * @private
 */
const LIMITS = [
    // Minute.
    { min: 0, max: 59 },
    // Heure.
    { min: 0, max: 23 },
    // Jour du mois.
    { min: 1, max: 31 },
    // Mois.
    { min: 1, max: 12 },
    // Jour de la semaine.
    { min: 0, max: 7 },
];
/* eslint-enable jsdoc/check-types */

/**
 * Le nombre maximum de jours pour chaque mois.
 *
 * @type {number[]}
 * @private
 */
const MAX_DAYS_IN_MONTHS = [31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31];

/**
 * Le message d'erreur (qui sera suivi de l'expression <em>cron</em>) renvoyé
 * dans l'exception.
 *
 * @type {string}
 * @private
 */
const ERROR = "Syntax error, unrecognized expression: ";

/**
 * La classe d'une expression <em>cron</em>.
 *
 * @class CronExp
 */
const CronExp = class {

    /**
     * Les valeurs possibles pour les minutes.
     *
     * @type {Field}
     * @private
     */
    _minutes;

    /**
     * Les valeurs possibles pour les heures.
     *
     * @type {Field}
     * @private
     */
    _hours;

    /**
     * Les valeurs possibles pour le jour du mois.
     *
     * @type {Field}
     * @private
     */
    _date;

    /**
     * Les valeurs possibles pour le mois (dont les numéros commencent à zéro
     * pour utiliser la même numérotation que
     * <code>Date.prototype.getMonth()</code>).
     *
     * @type {Field}
     * @private
     */
    _month;

    /**
     * Les valeurs possibles pour le jour de la semaine (en utilisant toujours
     * <code>0</code> pour le dimanche pour utiliser la même numéroration que
     * <code>Date.prototype.getDay()</code>).
     *
     * @type {Field}
     * @private
     */
    _day;

    /**
     * Crée une expression <em>cron</em>.
     *
     * @param {string} pattern Le motif de l'expression <em>cron</em>.
     * @throws {Error}      Si la syntaxe du motif est incorrecte.
     * @throws {RangeError} Si un intervalle est invalide (hors limite ou quand
     *                      la borne supérieure est plus petite que la borne
     *                      inférieure).
     * @throws {TypeError}  Si le constructeur est appelé sans le mot clé
     *                      <code>new</code> ou si le motif n'est pas une chaine
     *                      de caractères.
     */
    constructor(pattern) {
        // Remplacer l'éventuelle chaine spéciale par son équivalent et séparer
        // les cinq champs (minute, heure, jour du mois, mois et jour de la
        // semaine).
        const fields = (NICKNAMES[pattern] ?? pattern).split(/\s+/u);
        if (5 !== fields.length) {
            throw new Error(ERROR + pattern);
        }

        // Parcourir les cinq champs.
        const conds = fields.map((field, index) => {
            if ("*" === field) {
                return Field.all(LIMITS[index].min, LIMITS[index].max);
            }

            // Gérer les motifs "*/x".
            const result = (/^\*\/(?<step>\d+)$/u).exec(field);
            if (undefined !== result?.groups) {
                const step = Number.parseInt(result.groups.step, 10);
                if (0 === step) {
                    throw new RangeError(ERROR + pattern);
                }

                return Field.range(LIMITS[index].min, LIMITS[index].max, step);
            }

            // Gérer le nom des mois ("jan", "feb", ...) et des jours de la
            // semaine ("sun", "mon", ...).
            const value = NAMES[index].indexOf(field.toLowerCase());
            if (-1 !== value) {
                return Field.of(value);
            }

            // Gérer les listes.
            return Field.flat(field.split(",").map((range) => {
                const subresult =
                    // eslint-disable-next-line unicorn/no-unsafe-regex
                    (/^(?<min>\d+)(?:-(?<max>\d+)(?:\/(?<step>\d+))?)?$/u)
                                                                   .exec(range);
                if (undefined === subresult?.groups) {
                    throw new Error(ERROR + pattern);
                }

                const min  = Number.parseInt(subresult.groups.min, 10);
                const max  = undefined === subresult.groups.max
                                    ? min
                                    : Number.parseInt(subresult.groups.max, 10);
                const step = undefined === subresult.groups.step
                                   ? 1
                                   : Number.parseInt(subresult.groups.step, 10);

                // Vérifier que les valeurs sont dans les intervalles.
                if (min < LIMITS[index].min || LIMITS[index].max < max ||
                        max < min || 0 === step) {
                    throw new RangeError(ERROR + pattern);
                }

                return Field.range(min, max, step);
            }));
        });

        this._minutes = conds[0];
        this._hours = conds[1];
        this._date = conds[2];
        // Faire commencer les mois à zéro pour janvier.
        this._month = conds[3].map((v) => v - 1);
        // Toujours utiliser zéro pour le dimanche.
        this._day = conds[4].map((v) => (7 === v ? 0 : v));

        // Récupérer le mois le plus long parmi tous les mois autorisés.
        const max = Math.max(...this._month.values()
                                           .map((m) => MAX_DAYS_IN_MONTHS[m]));
        if (max < this._date.min) {
            throw new RangeError(ERROR + pattern);
        }
    }

    /**
     * Teste si une date respecte l'expression.
     *
     * @param {Date} [date=new Date()] La date qui sera testée (ou l'instant
     *                                 présent par défaut).
     * @returns {boolean} <code>true</code> si l'expression est respectée ;
     *                    sinon <code>false</code>.
     */
    test(date = new Date()) {
        // Vérifier que la minute, l'heure et le mois respectent les conditions.
        if (!this._minutes.test(date.getMinutes()) ||
                !this._hours.test(date.getHours()) ||
                !this._month.test(date.getMonth())) {
            return false;
        }

        // Quand le jour du mois et le jour de la semaine sont renseignés
        // (différent de l'astérisque), vérifier si au moins une des deux
        // conditions est respectée.
        if (this._date.restricted && this._day.restricted) {
            return this._date.test(date.getDate()) ||
                   this._day.test(date.getDay());
        }

        // Sinon : soit le jour du mois, soit le jour de la semaine ou aucun des
        // deux ne sont renseignés. Dans ce cas, vérifier classiquement les deux
        // conditions.
        return this._date.test(date.getDate()) &&
               this._day.test(date.getDay());
    }

    /**
     * Calcule la prochaine date (ou garde la date de début si les minutes
     * respectent la condition) en vérifiant seulement la condition des minutes.
     *
     * @param {Date} start La date de début.
     * @returns {Date} La prochaine date vérifiant la condition des minutes.
     * @private
     */
    _nextMinutes(start) {
        if (this._minutes.test(start.getMinutes())) {
            return start;
        }

        const date = new Date(start.getTime());
        const next = this._minutes.next(date.getMinutes());
        if (undefined === next) {
            date.setHours(date.getHours() + 1);
            date.setMinutes(this._minutes.min);
        } else {
            date.setMinutes(next);
        }
        return date;
    }

    /**
     * Calcule la prochaine date (ou garde la date de début si les heures
     * respectent la condition) en vérifiant seulement la condition des heures.
     *
     * @param {Date} start La date de début.
     * @returns {Date} La prochaine date vérifiant la condition des heures.
     * @private
     */
    _nextHours(start) {
        if (this._hours.test(start.getHours())) {
            return start;
        }

        const date = new Date(start.getTime());
        date.setMinutes(this._minutes.min);
        const next = this._hours.next(date.getHours());
        if (undefined === next) {
            date.setDate(date.getDate() + 1);
            date.setHours(this._hours.min);
        } else {
            date.setHours(next);
        }
        return date;
    }

    /**
     * Calcule la prochaine date (ou garde la date de début si le jour du mois
     * respecte la condition) en vérifiant seulement la condition du jour du
     * mois.
     *
     * @param {Date} start La date de début.
     * @returns {Date} La prochaine date vérifiant la condition du jour du
     *                 mois.
     * @private
     */
    _nextDate(start) {
        if (this._date.test(start.getDate())) {
            return start;
        }

        const date = new Date(start.getTime());
        date.setHours(this._hours.min);
        date.setMinutes(this._minutes.min);
        const next = this._date.next(date.getDate());
        if (undefined === next ||
                next > new Date(date.getFullYear(),
                                date.getMonth() + 1,
                                0).getDate()) {
            date.setMonth(date.getMonth() + 1);
            date.setDate(this._date.min);
        } else {
            date.setDate(next);
        }
        return date;
    }

    /**
     * Calcule la prochaine date (ou garde la date de début si le jour de la
     * semaine respecte la condition) en vérifiant seulement la condition du
     * jour de la semaine.
     *
     * @param {Date} start La date de début.
     * @returns {Date} La prochaine date vérifiant la condition du jour de la
     *                 semaine.
     * @private
     */
    _nextDay(start) {
        if (this._day.test(start.getDay())) {
            return start;
        }

        const date = new Date(start.getTime());
        date.setHours(this._hours.min);
        date.setMinutes(this._minutes.min);
        const next = this._day.next(date.getDay()) ?? this._day.min;
        date.setDate(date.getDate() + (next + (7 - date.getDay())) % 7);
        return date;
    }

    /**
     * Calcule la prochaine date (ou garde la date de début si le jour du mois
     * ou de la semaine respecte la condition) en vérifiant seulement la
     * condition du jour du mois ou de la semaine.
     *
     * @param {Date} start La date de début.
     * @returns {Date} La prochaine date vérifiant la condition du jour du
     *                 mois ou de la semaine.
     * @private
     */
    _nextDateDay(start) {
        const nextDate = this._nextDate(start);
        const nextDay = this._nextDay(start);
        // Si le jour du mois et de la semaine sont renseignés (différent de
        // l'astérisque), prendre la prochaine date la plus proche (qui est
        // l'équivalent de prendre la prochaine date vérifiant au moins une des
        // deux conditions).
        // Sinon, prendre la prochaine date la plus éloignée (qui est
        // l'équivalent de prendre la prochaine date vérifiant les deux
        // conditions).
        return this._date.restricted && this._day.restricted
                    ? new Date(Math.min(nextDate.getTime(), nextDay.getTime()))
                    : new Date(Math.max(nextDate.getTime(), nextDay.getTime()));
    }

    /**
     * Calcule la prochaine date (ou garde la date de début si le mois respecte
     * la condition) en vérifiant seulement la condition du mois.
     *
     * @param {Date} start La date de début.
     * @returns {Date} La prochaine date vérifiant la condition du mois.
     * @private
     */
    _nextMonth(start) {
        if (this._month.test(start.getMonth())) {
            return start;
        }

        const date = new Date(start.getTime());
        // Mettre temporairement le premier jour du mois et calculer le bon jour
        // après avoir trouver le bon mois.
        date.setDate(1);
        date.setHours(this._hours.min);
        date.setMinutes(this._minutes.min);
        const next = this._month.next(date.getMonth());
        if (undefined === next) {
            date.setFullYear(date.getFullYear() + 1);
            date.setMonth(this._month.min);
        } else {
            date.setMonth(next);
        }

        return this._nextDateDay(date);
    }

    /**
     * Calcule la prochaine date respectant l'expression.
     *
     * @param {Date} [start=new Date()] La date de début (ou l'instant présent
     *                                  par défaut).
     * @returns {Date} La prochaine date respectant l'expression.
     */
    next(start = new Date()) {
        let date = new Date(start.getTime());
        date.setSeconds(0, 0);
        date.setMinutes(date.getMinutes() + 1);

        date = this._nextMinutes(date);
        date = this._nextHours(date);
        date = this._nextDateDay(date);
        date = this._nextMonth(date);
        return date;
    }
};

/**
 * @module
 * @author Sébastien Règne
 * @license MIT
 */

/**
 * La valeur maximale du delai accepté par <em>Node.js</em>.
 *
 * @type {number}
 * @private
 * @see https://nodejs.org/api/timers.html
 */
const MAX_DELAY = 2_147_483_647;

/**
 * La classe d'une tâche <em>cronée</em>.
 *
 * @class Cron
 */
const Cron = class {

    /**
     * La liste des expressions <em>cron</em> indiquant les horaires d'exécution
     * de la tâche.
     *
     * @type {CronExp[]}
     * @private
     */
    _cronexes;

    /**
     * La fonction appelée à chaque horaire indiqué dans les expressions
     * <em>cron</em>.
     *
     * @type {Function}
     * @private
     */
    _func;

    /**
     * Le <code>this</code> utilisé pour la fonction.
     *
     * @type {any}
     * @private
     */
    // eslint-disable-next-line no-invalid-this
    _thisArg = this;

    /**
     * La liste des paramètres passés à la fonction.
     *
     * @type {any[]}
     * @private
     */
    _args = [];

    /**
     * L'identifiant du minuteur de la prochaine exécution ; ou
     * <code>undefined</code> s'il la tâche est active, mais il n'y a pas de
     * prochaine exécution (car il y a aucune expression <em>cron</em>) ; ou
     * <code>null</code> si la tâche est désactivée.
     *
     * @type {any}
     * @private
     */
    _timeoutID = null;

    /**
     * Crée une tâche <em>cronée</em>.
     *
     * @param {string|string[]} cronex        La ou les expressions
     *                                        <em>cron</em> indiquant les
     *                                        horaires d'exécution de la tâche.
     *                                        Avec un tableau vide, la tâche ne
     *                                        sera jamais exécutée.
     * @param {Function}        func          La fonction appelée à chaque
     *                                        horaire indiqué dans les
     *                                        expressions <em>cron</em>.
     * @param {boolean}         [active=true] <code>true</code> (par défaut)
     *                                        pour activer la tâche ; sinon
     *                                        <code>false</code>.
     * @throws {Error}      Si la syntaxe d'une expession <em>cron</em> est
     *                      incorrecte.
     * @throws {RangeError} Si un intervalle d'une expression <em>cron</em> est
     *                      invalide (hors limite ou quand la borne supérieure
     *                      est plus petite que la borne inférieure).
     * @throws {TypeError}  Si le constructeur est appelé sans le mot clé
     *                      <code>new</code> ou si un des paramètres n'a pas le
     *                      bon type.
     */
    constructor(cronex, func, active = true) {
        const cronexes = Array.isArray(cronex) ? cronex
                                               : [cronex];
        this._cronexes = cronexes.map((p) => new CronExp(p));
        this._func = func;

        if (active) {
            this._schedule();
        }
    }

    /**
     * Récupère l'état de la tâche (active ou non).
     *
     * @returns {boolean} <code>true</code> si la tâche est active ; sinon
     *                    <code>false</code>.
     */
    get active() {
        return null !== this._timeoutID;
    }

    /**
     * Définit le <code>this</code> et les paramètres passés à la fonction.
     *
     * @param {any}    thisArg Le <code>this</code> utilisé pour la fonction.
     * @param {...any} args    Les paramètres passés à la fonction.
     * @returns {Cron} La tâche elle-même.
     */
    bind(thisArg, ...args) {
        this._thisArg = thisArg;
        this._args = args;
        return this;
    }

    /**
     * Remet les valeurs par défaut pour le <code>this</code> et les paramètres
     * passés à la fonction.
     *
     * @returns {Cron} La tâche elle-même.
     */
    unbind() {
        this._thisArg = this;
        this._args = [];
        return this;
    }

    /**
     * Définit les paramètres passés à la fonction.
     *
     * @param {...any} args Les paramètres passés à la fonction.
     * @returns {Cron} La tâche elle-même.
     */
    withArguments(...args) {
        this._args = args;
        return this;
    }

    /**
     * Exécute manuellement la fonction.
     */
    run() {
        this._func.bind(this._thisArg)(...this._args);
    }

    /**
     * Programme la prochaine exécution.
     *
     * @private
     */
    _schedule() {
        const next = this.next();
        if (undefined === next) {
            // Ne pas planifier de prochaine exécution, mais définir une valeur
            // différente de null pour l'identifiant du minuteur afin d'indiquer
            // que la tâche est active.
            this._timeoutID = undefined;
        } else {
            const delay = next.getTime() - Date.now();
            // eslint-disable-next-line unicorn/prefer-ternary
            if (MAX_DELAY >= delay) {
                // Planifier la prochaine exécution.
                this._timeoutID = setTimeout(() => {
                    this._schedule();
                    this.run();
                }, delay);
            } else {
                // Planifier des étapes intermédiaires car Node.js n'accepte pas
                // un grand délai.
                this._timeoutID = setTimeout(() => {
                    this._schedule();
                }, MAX_DELAY);
            }
        }
    }

    /**
     * Annule les prochaines exécutions.
     *
     * @private
     */
    _cancel() {
        clearTimeout(this._timeoutID);
        this._timeoutID = null;
    }

    /**
     * Active la tâche.
     *
     * @returns {boolean} <code>true</code> quand la tâche a été activée ;
     *                    <code>false</code> si elle était déjà active.
     */
    start() {
        if (this.active) {
            return false;
        }
        this._schedule();
        return true;
    }

    /**
     * Désactive la tâche.
     *
     * @returns {boolean} <code>true</code> quand la tâche a été désactivée ;
     *                    <code>false</code> si elle était déjà inactive.
     */
    stop() {
        if (!this.active) {
            return false;
        }
        this._cancel();
        return true;
    }

    /**
     * Teste si une date respecte une des expressions <em>cron</em> de la tâche.
     *
     * @param {Date} [date=new Date()] La date qui sera testée (ou l'instant
     *                                 présent par défaut).
     * @returns {boolean} <code>true</code> si une des expressions est
     *                    respectée ; sinon <code>false</code>.
     */
    test(date = new Date()) {
        return this._cronexes.some((c) => c.test(date));
    }

    /**
     * Calcule la prochaine date respectant une des expressions <em>cron</em> de
     * la tâche.
     *
     * @param {Date} [start=new Date()] La date de début (ou l'instant présent
     *                                  par défaut).
     * @returns {Date|undefined} La prochaine date respectant une des
     *                           expressions ou <code>undefined</code> s'il n'y
     *                           a pas de prochaine date (quand il y a aucune
     *                           expression <em>cron</em>).
     */
    next(start = new Date()) {
        if (0 === this._cronexes.length) {
            return undefined;
        }

        return new Date(Math.min(
            ...this._cronexes.map((c) => c.next(start).getTime()),
        ));
    }
};

export { Cron, CronExp };
